from typing import Literal
import numpy as np


def psi(expected: np.ndarray, actual: np.ndarray, bucket_type: Literal["bins", "quantiles"] = "quantiles", n_bins: int = 10) -> float:
    """Calculate PSI metric for two arrays.

    Parameters
    ----------
        expected : list-like
            Array of expected values
        actual : list-like
            Array of actual values
        bucket_type : str
            Binning strategy. Accepts two options: 'bins' and 'quantiles'. Defaults to 'bins'.
            'bins': input arrays are splitted into bins with equal
                and fixed steps b
                ased on 'expected' array
            'quantiles': input arrays are binned according to 'expected' array
                with given number of n_bins
        n_bins : int
            Number of buckets for binning. Defaults to 10.

    Returns
    -------
        A single float number
    """
    if bucket_type == "bins":
        min_val = expected.min()
        max_val = expected.max()
        bins = np.linspace(min_val, max_val, n_bins + 1)
    elif bucket_type == "quantiles":
        percentage = np.arange(0, n_bins + 1) / (n_bins) * 100
        bins = np.percentile(expected, percentage)
    # Calculate frequencies
    expected_percents = np.histogram(expected, bins)[0] / len(expected)
    actual_percents = np.histogram(actual, bins)[0] / len(actual)
    # Clip freaquencies to avoid zero division
    expected_percents = np.clip(expected_percents, a_min=0.00001, a_max=None)
    actual_percents = np.clip(actual_percents, a_min=0.00001, a_max=None)
    # Calculate PSI
    psi_value = (expected_percents - actual_percents) * np.log(expected_percents / actual_percents)
    psi_value = sum(psi_value)
    return psi_value





from typing import List

import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler


def decribe_risk_score(
    score_df: pd.DataFrame, pred_col: str, label_col: str,
    segments: List[float] = [
        .01, .025, .05, .075, .1, .125, .15,
        .2, .3, .4, .5, .6, .8, .9, 1
    ]
) -> pd.DataFrame:
    segments.sort()

    def get_segment(score: float):
        for s in segments:
            if score <= s:
                return s
        return None

    # creating propensity score based segments
    prop_df = score_df[[pred_col, label_col]]
    prop_df.columns = ["probability", "default_next_month"]
    prop_df['proba_segment'] = prop_df['probability'].apply(get_segment)
    summary = prop_df.groupby('proba_segment').agg(
        client=("default_next_month", "count"),
        bad=("default_next_month", lambda s: (s == 1).sum()),
        good=("default_next_month", lambda s: (s == 0).sum()),
    )
    summary["cumsum_client"] = summary["client"].cumsum()
    summary["cumsum_bad"] = summary["bad"].cumsum()
    summary['cumsum_good'] = summary['good'].cumsum()
    summary["%bad vs total bad"] = round((summary["bad"]/summary["cumsum_bad"].max()) * 100, 2)
    summary["%bad cumsum"] = summary["%bad vs total bad"].cumsum()
    summary["%good vs total good"] = round(summary["good"]/summary["cumsum_good"].max() * 100, 2)
    summary["%good cumsum"] = summary["%good vs total good"].cumsum()
    summary["def_rate%"] = round((summary['cumsum_bad'] / summary['cumsum_client']) * 100, 2)
    summary["approval_rate%"] = round(summary["cumsum_client"] / len(prop_df) * 100, 2)
    summary['FPR'] = 1-summary['cumsum_good'] / (summary['cumsum_good'].max())
    summary['TPR'] = 1-summary['cumsum_bad'] / (summary['cumsum_bad'].max())
    summary[["bad_scaled", "good_scaled"]] = np.round(MinMaxScaler().fit_transform(summary[["bad", "good"]]), 3)
    summary = summary.reset_index()
    summary["band"] = [f"B{i}" for i in range(len(summary), 0, -1)]

    return summary[[
        'band', 'proba_segment', 'def_rate%', 'approval_rate%',
        'client', 'bad', 'good',
        '%bad vs total bad', '%bad cumsum',
        '%good vs total good',  '%good cumsum',
        'FPR', 'TPR', 'bad_scaled', 'good_scaled',
        'cumsum_client', 'cumsum_bad', 'cumsum_good',
    ]]
